@@ -294,6 +294,7 @@ class config(object):
 			self.configlist = [
 				self.configdict['env.d'],
 				self.configdict['repo'],
+				self.configdict['features'],
 				self.configdict['pkginternal'],
 				self.configdict['globals'],
 				self.configdict['defaults'],
@@ -461,13 +462,16 @@ class config(object):
 			# back up our incremental variables:
 			self.configdict={}
 			self._use_expand_dict = {}
-			# configlist will contain: [ env.d, globals, defaults, conf, pkg, backupenv, env ]
+			# configlist will contain: [ env.d, globals, features, defaults, conf, pkg, backupenv, env ]
 			self.configlist.append({})
 			self.configdict["env.d"] = self.configlist[-1]
 
 			self.configlist.append({})
 			self.configdict["repo"] = self.configlist[-1]
 
+			self.configlist.append({})
+			self.configdict["features"] = self.configlist[-1]
+
 			self.configlist.append({})
 			self.configdict["pkginternal"] = self.configlist[-1]
 
@@ -868,7 +872,7 @@ class config(object):
 			# reasonable defaults; this is important as without USE_ORDER,
 			# USE will always be "" (nothing set)!
 			if "USE_ORDER" not in self:
-				self["USE_ORDER"] = "env:pkg:conf:defaults:pkginternal:repo:env.d"
+				self["USE_ORDER"] = "env:pkg:conf:defaults:pkginternal:features:repo:env.d"
 				self.backup_changes("USE_ORDER")
 
 			if "CBUILD" not in self and "CHOST" in self:
@@ -1120,7 +1124,7 @@ class config(object):
 		archlist = self.archlist()
 		if not archlist:
 			writemsg(_("--- 'profiles/arch.list' is empty or "
-				"not available. Empty portage tree?\n"), noiselevel=1)
+				"not available. Empty ebuild repository?\n"), noiselevel=1)
 		else:
 			for group in groups:
 				if group not in archlist and \
@@ -1292,6 +1296,7 @@ class config(object):
 			del self._penv[:]
 			self.configdict["pkg"].clear()
 			self.configdict["pkginternal"].clear()
+			self.configdict["features"].clear()
 			self.configdict["repo"].clear()
 			self.configdict["defaults"]["USE"] = \
 				" ".join(self.make_defaults_use)
@@ -1452,6 +1457,8 @@ class config(object):
 		cp = cpv_getkey(mycpv)
 		cpv_slot = self.mycpv
 		pkginternaluse = ""
+		pkginternaluse_list = []
+		feature_use = []
 		iuse = ""
 		pkg_configdict = self.configdict["pkg"]
 		previous_iuse = pkg_configdict.get("IUSE")
@@ -1507,13 +1514,12 @@ class config(object):
 				cpv_slot = self.mycpv
 			else:
 				cpv_slot = pkg
-			pkginternaluse = []
 			for x in iuse.split():
 				if x.startswith("+"):
-					pkginternaluse.append(x[1:])
+					pkginternaluse_list.append(x[1:])
 				elif x.startswith("-"):
-					pkginternaluse.append(x)
-			pkginternaluse = " ".join(pkginternaluse)
+					pkginternaluse_list.append(x)
+			pkginternaluse = " ".join(pkginternaluse_list)
 
 		eapi_attrs = _get_eapi_attrs(eapi)
 
@@ -1590,6 +1596,9 @@ class config(object):
 			# regenerate() call in order to ensure that self.features
 			# is accurate.
 			has_changed = True
+			# Prevent stale features USE from corrupting the evaluation
+			# of USE conditional RESTRICT.
+			self.configdict["features"]["USE"] = ""
 
 		self._penv = []
 		cpdict = self._penvdict.get(cp)
@@ -1650,6 +1659,57 @@ class config(object):
 		if has_changed:
 			self.reset(keeping_pkg=1)
 
+		if explicit_iuse is None:
+			explicit_iuse = frozenset(x.lstrip("+-") for x in iuse.split())
+		if eapi_attrs.iuse_effective:
+			iuse_implicit_match = self._iuse_effective_match
+		else:
+			iuse_implicit_match = self._iuse_implicit_match
+
+		if pkg is None:
+			raw_restrict = pkg_configdict.get("RESTRICT")
+		else:
+			raw_restrict = pkg._raw_metadata["RESTRICT"]
+
+		restrict_test = False
+		if raw_restrict:
+			try:
+				if built_use is not None:
+					restrict = use_reduce(raw_restrict,
+						uselist=built_use, flat=True)
+				else:
+					restrict = use_reduce(raw_restrict,
+						uselist=frozenset(x for x in self['USE'].split()
+						if x in explicit_iuse or iuse_implicit_match(x)),
+						flat=True)
+			except PortageException:
+				pass
+			else:
+				restrict_test = "test" in restrict
+
+		pkginternaluse_before = pkginternaluse
+		if "test" in self.features:
+			# This is independent of IUSE and RESTRICT, so that the same
+			# value can be shared between packages with different settings,
+			# which is important when evaluating USE conditional RESTRICT
+			# above.
+			feature_use.append("test")
+
+			if restrict_test:
+				# Handle it like IUSE="-test", since features USE is
+				# independent of RESTRICT.
+				pkginternaluse_list.append("-test")
+				pkginternaluse = " ".join(pkginternaluse_list)
+				self.configdict["pkginternal"]["USE"] = pkginternaluse
+
+		feature_use = " ".join(feature_use)
+		if (feature_use != self.configdict["features"].get("USE", "") or
+			pkginternaluse is not pkginternaluse_before):
+			self.configdict["features"]["USE"] = feature_use
+			# TODO: can we avoid that?
+			self.reset(keeping_pkg=1)
+			has_changed = True
+
 		env_configdict = self.configdict['env']
 
 		# Ensure that "pkg" values are always preferred over "env" values.
@@ -1677,11 +1737,8 @@ class config(object):
 		# package has different IUSE.
 		use = set(self["USE"].split())
 		unfiltered_use = frozenset(use)
-		if explicit_iuse is None:
-			explicit_iuse = frozenset(x.lstrip("+-") for x in iuse.split())
 
 		if eapi_attrs.iuse_effective:
-			iuse_implicit_match = self._iuse_effective_match
 			portage_iuse = set(self._iuse_effective)
 			portage_iuse.update(explicit_iuse)
 			if built_use is not None:
@@ -1693,7 +1750,6 @@ class config(object):
 			self.configdict["pkg"]["IUSE_EFFECTIVE"] = \
 				" ".join(sorted(portage_iuse))
 		else:
-			iuse_implicit_match = self._iuse_implicit_match
 			portage_iuse = self._get_implicit_iuse()
 			portage_iuse.update(explicit_iuse)
 
@@ -1701,49 +1757,21 @@ class config(object):
 		self.configdict["env"].addLazySingleton(
 			"PORTAGE_IUSE", _lazy_iuse_regex, portage_iuse)
 
-		if pkg is None:
-			raw_restrict = pkg_configdict.get("RESTRICT")
-		else:
-			raw_restrict = pkg._raw_metadata["RESTRICT"]
-
-		restrict_test = False
-		if raw_restrict:
-			try:
-				if built_use is not None:
-					restrict = use_reduce(raw_restrict,
-						uselist=built_use, flat=True)
-				else:
-					# Use matchnone=True to ignore USE conditional parts
-					# of RESTRICT, since we want to know whether to mask
-					# the "test" flag _before_ we know the USE values
-					# that would be needed to evaluate the USE
-					# conditionals (see bug #273272).
-					restrict = use_reduce(raw_restrict,
-						matchnone=True, flat=True)
-			except PortageException:
-				pass
-			else:
-				restrict_test = "test" in restrict
-
 		ebuild_force_test = not restrict_test and \
 			self.get("EBUILD_FORCE_TEST") == "1"
 
 		if "test" in explicit_iuse or iuse_implicit_match("test"):
-			if "test" not in self.features:
-				use.discard("test")
-			elif restrict_test or \
+			if "test" in self.features:
+				if ebuild_force_test and "test" in self.usemask:
+					self.usemask = \
+						frozenset(x for x in self.usemask if x != "test")
+			if restrict_test or \
 				("test" in self.usemask and not ebuild_force_test):
 				# "test" is in IUSE and USE=test is masked, so execution
 				# of src_test() probably is not reliable. Therefore,
 				# temporarily disable FEATURES=test just for this package.
 				self["FEATURES"] = " ".join(x for x in self.features \
 					if x != "test")
-				use.discard("test")
-			else:
-				use.add("test")
-				if ebuild_force_test and "test" in self.usemask:
-					self.usemask = \
-						frozenset(x for x in self.usemask if x != "test")
 
 		if eapi_attrs.feature_flag_targetroot and \
 			("targetroot" in explicit_iuse or iuse_implicit_match("targetroot")):
@@ -1900,12 +1928,6 @@ class config(object):
 		iuse_implicit.add("build")
 		iuse_implicit.add("bootstrap")
 
-		# Controlled by FEATURES=test. Make this implicit, so handling
-		# of FEATURES=test is consistent regardless of explicit IUSE.
-		# Users may use use.mask/package.use.mask to control
-		# FEATURES=test for all ebuilds, regardless of explicit IUSE.
-		iuse_implicit.add("test")
-
 		return iuse_implicit
 
 	def _getUseMask(self, pkg, stable=None):
